<?php
/**
 * Template Logic
 * @package     Template
 * @copyright   Copyright (C) 2011 MicroEngine Inc.
 */
class Template_Logic
{
    /** システム文字コード */
    const CHAR_CODE = 'UTF-8';
    /** 文字コード検出順 */
    const DETECT_ENCODING = 'UTF-8, SJIS, EUC-JP, ISO-2022-JP, ISO-8859-1';
    /** テンプレートタイプ：シングルタイプ */
    const TYPE_SINGLE = 'single';
    /** テンプレートタイプ：マルチタイプ */
    const TYPE_MULTI = 'multi';
    /** テンプレートタイプ：サブタイプ */
    const TYPE_SUB = 'sub';
    /** 配信方式：静的 */
    const OUTPUT_STATIC = 'static';
    /** 配信方式：動的 */
    const OUTPUT_DYNAMIC = 'dynamic';
    /** テンプレートの１行PHPが呼び出すファイルのCMS_PATHからのパス */
    const SITE_FILE_PATH = '/app/site.php';
    /** TEMPLATEディレクトリ */
    const TEMPLATE_DIR = '/template';

    /**
     * DAOオブジェクト
     * @var Template_Dao
     */
    private $dao;

    /**
     * コンストラクタ
     */
    public function __construct()
    {
        $this->dao = new Template_Dao();
    }

    /**
     * SELECT
     * 1レコードのオブジェクトを返す
     * @param string $_id _idフィールド値
     * @param string $class_name クラス名
     * @return object $obj
     */
    public function select($_id, $class_name = null)
    {
        return $this->dao->select($_id, $class_name);
    }

    /**
     * テンプレートID取得
     * @param string $template_id
     * @return array $rows
     */
    public function select_sub_template_list($template_id)
    {
        return $this->dao->select_sub_template_list($template_id);
    }

    /**
     * テンプレート登録
     * @param Template_Form $form
     * @return boolean
     */
    public function register_template(Template_Form $form)
    {
        if (!$this->validation($form)) {
            return false;
        }

        $this->output_template($form);

        if (empty($form->_id)) {
            // 新規追加
            $this->dao->insert($form);
            // ページタイプをアップデート
            $page_dao = new Page_Dao();
            $page = $page_dao->select($form->page_id);
            $page->page_type = Page_Logic::PAGE_TYPE_TEMPLATE;
            $page->template_id = $form->_id;
            $page_dao->update($page);
        } else {
            // 編集
            $this->dao->update($form);
        }
        // 参照関係を登録
        $this->register_relation($form);

        return true;
    }

    /**
     * テンプレートフォーム取得
     * @param  object $page
     * @return Template_Form $form
     */
    public function get_template_form_by_page($page)
    {
        if (empty($page->template_id)) {
            $form = new Template_Form();
            $form->template_path = $page->page_path;
        } else {
            $form = $this->dao->select($page->template_id, 'Template_Form');
        }
        $form->page_id = $page->_id;
        if (empty($form->template_type)) {
            $form->template_type = self::TYPE_SINGLE;
        }
        if (empty($form->output_method)) {
            $form->output_method = self::OUTPUT_STATIC;
        }
        $form->page_path = $page->page_path;
        return $form;
    }

    /**
     * テンプレートフォーム取得
     * @param string $page_id
     * @return Template_Form $form
     */
    public function get_edit_template_form($page_id)
    {
        // ページ情報取得
        $page_dao = new Page_Dao();
        $page = $page_dao->select($page_id);

        // テンプレート情報取得
        $form = $this->get_template_form_by_page($page);

        // ソースコード取得
        $this->get_template_file($form);

        // ソースコードの実態参照がtextareaでデコードされないように、'&'を変換して２重エンコード状態にする
        $form->source_code = str_replace('&', '&amp;', $form->source_code);

        // 予約MEタグ挿入
        $parser = new Template_Parser();
        $form->document = $parser->parse($form->source_code);
        $this->set_reserved_metag($form);

        return $form;
    }
    
    /**
     * ページ書き出し（テンプレート単位の再構築機能）
     * 指定したテンプレートから作成されたページを一括して書きだす。
     * @param string $template_id 
     */
    public function write_page($template_id)
    {
        // ページID取得
        $page_dao = new Page_Dao();
        $page_list = $page_dao->select_page_id_by_template_id($template_id);
         
        // 関連ページ書き出し
        $page_output = new Page_Output();
        foreach ($page_list as $rel_page_row) {
            // $page_outputは再利用する
            $page_output->output($rel_page_row['_id']);
        }
    }

    /**
     * 予約MEタグ挿入
     * @param Template_Form $form
     */
    private function set_reserved_metag(Template_Form $form)
    {
        // titleタグ
        $title_tag = $form->document->get_element_by_id('_title');
        if ($title_tag === null) {
            if (preg_match('!(<title.*?>)(.*?)(</title>)!i', $form->source_code, $matches)) {
                $title = Me_util::html_escape($matches[2]);
                //タグがあれば、置換する。
                $form->source_code = preg_replace(
                        '!(<title.*?>)(.*?)(</title>)!i',
                        "$1<!--{text id=\"_title\" ref_id=\"_title\" initial_value=\"" . $title . "\"}-->$2<!--{/text}-->$3",
                        $form->source_code
                );
            }
        }
        // metaタグ
        $this->set_reserved_metag_to_meta($form, 'keywords');
        $this->set_reserved_metag_to_meta($form, 'description');
    }

    /**
     * 予約MEタグをmetaタグに挿入
     * @param Template_Form $form
     * @param string $meta_name
     */
    private function set_reserved_metag_to_meta(Template_Form $form, $meta_name)
    {
        // metaタグ
        $metag = $form->document->get_element_by_id('_' . $meta_name);
        if ($metag === null) {
            if (preg_match('!(<meta.*?name=["\'])' . $meta_name . '(["\'].*?>)!i', $form->source_code, $matches)) {
                $meta = $matches[0];
                $content_value = '';
                if (preg_match('/content=(""|".*?[^\\\\]"|\'\'|\'.*?[^\\\\]\')/', $meta, $match_content)) {
                    $content_value = Me_Util::html_escape(substr($match_content[1], 1, strlen($match_content[1]) - 2));
                }
                //タグがあれば、置換する。
                $form->source_code = preg_replace(
                        '!(<meta.*?name=["\'])' . $meta_name . '(["\'].*?>)!i',
                        "<!--{attr id=\"_{$meta_name}\" target=\"content\" initial_value=\"" . $content_value . "\"}-->$0<!--{/attr}-->",
                        $form->source_code
                );
            }
        }
    }

    /**
     * テンプレートファイルの内容を取得
     * 文字コードはUTF-8に変換する
     * @param Template_Form $form
     * @param string $char_code 文字コード
     */
    public function get_template_file(Template_Form $form, $char_code = null)
    {
        // ファイルの内容取得
        if (empty($form->_id)) {
            $form->source_code = Me_Util::get_file(SITE_ROOT . $form->template_path);
        } else {
            $form->source_code = Me_Util::get_file(DATA_ROOT . self::TEMPLATE_DIR . $form->template_path);
        }

        if (!empty($char_code)) {
            $form->char_code = $char_code;
        } elseif (empty($form->char_code)) {
            // 文字コードが指定されていない場合は、mb関数で取得する。
            $form->char_code = mb_detect_encoding($form->source_code, self::DETECT_ENCODING);
        }

        // 文字コードがUTF-8以外の場合は変換する
        if (!empty($form->char_code) && $form->char_code != self::CHAR_CODE) {
            $form->source_code = mb_convert_encoding(
                    $form->source_code, self::CHAR_CODE, $form->char_code
            );
        }
    }

    /**
     * 入力値チェック
     * @param Template_Form $form
     * @return boolean
     */
    public function validation(Template_Form $form)
    {
        // エラーメッセージを初期化
        $form->errors = array();

        // テンプレートタイプ
        // 必須チェック
        if (strlen($form->template_type) < 1) {
            $form->errors['template_type'] = 'テンプレートタイプを選択してください';
        }
        // 配信方式
        // 必須チェック
        if (strlen($form->output_method) < 1) {
            $form->errors['output_method'] = '配信方式を選択してください';
        }
        // 文字コード
        // 必須チェック
        if (strlen($form->char_code) < 1) {
            $form->errors['char_code'] = '文字コードを入力してください';
        }

        // 備考チェック
        if (50 < mb_strlen($form->note)) {
            $form->errors['note'] = '50文字以内で入力してください';
        }

        // エラー有無確認
        if (!empty($form->errors)) {
            $form->errors['MESSAGE'] = 'エラーが発生しました。';
            return false;
        }
        return true;
    }

    /**
     * テンプレート削除
     * @param string $template_id
     * @return boolean
     */
    public function delete_template($template_id)
    {
        $template = $this->dao->select($template_id);

        // テンプレート削除前に静的配信に変更
        $template->output_method = self::OUTPUT_STATIC;
        $this->dao->update($template);
        // 関連ページを書きだす
        $this->output_page($template_id);
        // 関連ページ情報を更新する。
        $page_dao = new Page_Dao();
        $page_dao->delete_not_public($template_id);
        $page_dao->update_page_type($template_id);

        // DBから削除
        $this->dao->delete($template);

        // Relationテーブルから該当テンプレートのレコードを削除
        $relation_dao = new Relation_Dao();
        $relation_dao->delete_from_template($template_id);

        // テンプレートファイルを削除
        $file_path = DATA_ROOT . self::TEMPLATE_DIR . $template->template_path;
        if (file_exists($file_path)) {
            unlink($file_path);
        }

        return true;
    }

    /**
     * テンプレート関連ページ出力
     * @param string $template_id
     */
    private function output_page($template_id)
    {
        // 関連ページ取得
        $page_dao = new Page_Dao();
        $page_list = $page_dao->select_page_list_by_template_id($template_id);

        // 関連ページ書き出し
        $page_output = new Page_Output();
        foreach ($page_list as $page_row) {
            // $page_outputは再利用する
            $page_output->output($page_row['_id']);
        }
    }

    /**
     * テンプレート関連ファイル出力
     * @param Template_Form $form テンプレートフォーム
     */
    private function output_template(Template_Form $form)
    {
        // document構築
        $this->create_document($form);

        // テンプレート関連ファイル書き出し
        $this->write_template($form);
    }

    /**
     * テンプレート書き出し
     * 書き出すパスにファイルが存在する場合は上書きする。
     * @param Template_Form $form テンプレートフォーム
     * @return boolean
     */
    private function write_template(Template_Form $form)
    {
        // 書き出すファイルパス
        $file_path = DATA_ROOT . self::TEMPLATE_DIR . $form->template_path;

        // 文字コードがUTF-8以外の場合は、元の文字コードに戻す。
        if (!empty($form->char_code) && $form->char_code != self::CHAR_CODE) {
            $form->source_code = mb_convert_encoding(
                    $form->source_code,
                    $form->char_code,
                    self::CHAR_CODE
            );
        }

        // ファイル書き出し
        Me_Util::write_file($file_path, $form->source_code);
        
        // サブテンプレートの場合は、公開側のファイルを削除する。
        if ($form->template_type === self::TYPE_SUB) {
            $file_path = SITE_ROOT . $form->template_path;
            if (file_exists($file_path)) {
                unlink($file_path);
            }
        }

        return true;
    }

    /**
     * リレーション登録
     * @param Template_Form $form
     */
    private function register_relation(Template_Form $form)
    {
        $relation_list = array();
        if (!empty($form->document->ref_page_list)) {
            foreach ($form->document->ref_page_list as $ref_page) {
                $relation = new Relation();
                $relation->from_template = $form->_id;
                $relation->ref_page = $ref_page;
                $relation_list[] = $relation;
            }
        }
        if (!empty($form->document->ref_template_list)) {
            foreach ($form->document->ref_template_list as $ref_template) {
                $relation = new Relation();
                $relation->from_template = $form->_id;
                $relation->ref_template = $ref_template;
                $relation_list[] = $relation;
            }
        }
        $relation_dao = new Relation_Dao();
        $relation_dao->register_array($relation_list, $form->_id);
    }

    /**
     * DOM構築
     * @param Template_Form $form
     */
    private function create_document(Template_Form $form)
    {
        $parser = new Template_Parser();
        $parser->output_method = $form->output_method;
        $form->document = $parser->parse($form->source_code);
    }

    /**
     * preview用head情報更新
     * baseタグを制御
     * @param Template_Form $form
     */
    public function update_preview_header(Template_Form $form)
    {
        // 追加するタグの内容
        $scheme = (empty($_SERVER['HTTPS'])) ? 'http' : 'https';
        $add_tag = '<base href="' . $scheme . '://' . $_SERVER['HTTP_HOST']
            . SITE_PATH . $form->template_path . '">';

        // baseタグを追加する。
        if (preg_match("/<head.*?>/i", $form->source_code, $matches)) {
            // <head>が存在すれば、その後にタグを追加する。
            $form->source_code = preg_replace(
                    '|' . $matches[0] . '|',
                    $matches[0] . $add_tag,
                    $form->source_code
            );
        } else if (preg_match("/<html.*?>/i", $html, $matches)) {
            // <html>が存在すれば、その後にタグを追加する。
            $form->source_code = preg_replace(
                    '/' . $matches[0] . '/',
                    $matches[0] . '<head>' . $add_tag . '</head>',
                    $form->source_code
            );
        } else {
            // headタグもhtmlタグも存在しなければ、先頭に追加する。
            $form->source_code = "<head>" . $add_tag . "</head>" . $form->source_code;
        }
    }

    /**
     * ページタイトル取得
     * @param boolean $is_edit
     * @return string
     */
    public function get_page_title($is_edit)
    {
        return ($is_edit) ? 'テンプレート編集' : 'テンプレート新規作成';
    }

    /**
     * プレビュー（インスペクタ）
     * @param Template_Form $form
     */
    public function preview(Template_Form $form)
    {
        $page_logic = new Page_Logic();
        $page_logic->update_header($form);
    }
}